from zope.interface import implements

from ftw.journal.interfaces import IJournalEntryEvent


class JournalEntryEvent(object):
    """Event for journal entries"""
    implements(IJournalEntryEvent)
    
    def __init__(self, obj, comment, action=None, actor=None, time=None):
        self.object = obj
        self.action = action
        self.comment = comment
        self.actor = actor
        self.time = time



from DateTime import DateTime
from ftw.journal.config import JOURNAL_ENTRIES_ANNOTATIONS_KEY
from ftw.journal.events.events import JournalEntryEvent
from ftw.journal.interfaces import \
    IAnnotationsJournalizable, IWorkflowHistoryJournalizable
from ftw.journal.testing import FTW_JORUNAL_INTEGRATION_TESTING
from plone.app.testing import TEST_USER_ID
from zope.annotation.interfaces import IAnnotations
from zope.event import notify
from zope.interface import alsoProvides
import unittest2 as unittest
from Products.CMFCore.utils import getToolByName


class TestFtwJournalIntegration(unittest.TestCase):

    layer = FTW_JORUNAL_INTEGRATION_TESTING

    def test_event_annotation(self):
        """ Test the event with IAnnotationsJournalizable interface
        """

        portal = self.layer['portal']
        folder = portal[portal.invokeFactory('Folder', 'f1', )]
        doc = folder[folder.invokeFactory('Document', 'd1', )]

        alsoProvides(doc, IAnnotationsJournalizable)

        notify(JournalEntryEvent(doc, "added new line", "modified"))

        journal = IAnnotations(doc).get(JOURNAL_ENTRIES_ANNOTATIONS_KEY)[0]

        self.assertTrue('modified' in journal.get('action'))
        self.assertTrue('added new line' in journal.get('comments'))
        self.assertTrue(TEST_USER_ID in journal.get('actor'))
        self.assertTrue(DateTime().Date() in journal.get('time').Date())

    def test_event_workflow_history(self):
        """ Test the event with IWorkflowHistoryJournalizable interface
        """

        portal = self.layer['portal']
        folder = portal[portal.invokeFactory('Folder', 'f1', )]
        doc = folder[folder.invokeFactory('Document', 'd1', )]

        wftool = getToolByName(doc, 'portal_workflow')
        wftool.manage_changeWorkflows('simple_publication_workflow')

        alsoProvides(doc, IWorkflowHistoryJournalizable)

        notify(JournalEntryEvent(doc, "new workflow", "modified"))

        journal = wftool.getStatusOf('simple_publication_workflow', doc)

        self.assertTrue('modified' in journal.get('action'))
        self.assertTrue('new workflow' in journal.get('comments'))
        self.assertTrue('private' in journal.get('review_state'))
        self.assertTrue(TEST_USER_ID in journal.get('actor'))
        self.assertTrue(DateTime().Date() in journal.get('time').Date())

    def test_journal_is_none_if_no_workflow_defined(self):
        """ Test the event with IWorkflowHistoryJournalizable interface
        """

        portal = self.layer['portal']
        folder = portal[portal.invokeFactory('Folder', 'f1', )]
        doc = folder[folder.invokeFactory('Document', 'd1', )]

        alsoProvides(doc, IWorkflowHistoryJournalizable)

        notify(JournalEntryEvent(doc, "new workflow", "modified"))

        wftool = getToolByName(doc, 'portal_workflow')
        journal = wftool.getStatusOf('simple_publication_workflow', doc)
        self.assertIsNone(journal)



from Acquisition import aq_inner
from plone.protect import utils
from Products.CMFCore.utils import getToolByName


class WorkflowHistoryJournalizable(object):
    """Adapter to create journal entries in the workflow history"""

    def __init__(self, context):
        self.context = aq_inner(context)

    def __call__(self, action, comment, actor, time):
        context = self.context
        workflow_tool = getToolByName(context, 'portal_workflow')

        workflows = workflow_tool.getWorkflowsFor(context)

        if not workflows:
            return

        workflow_id = workflows[0].id
        review_state = workflow_tool.getInfoFor(context, 'review_state', None)

        history_entry = {
                         'action': action,
                         'review_state': review_state,
                         'comments': comment,
                         'actor': actor,
                         'time': time,
                         }

        # If we have plone.protect > 3.0, mark the journal write as safe
        if 'safeWrite' in dir(utils):
            utils.safeWrite(context)
            utils.safeWrite(review_state)

        workflow_tool.setStatusOf(workflow_id, context, history_entry)



from zope.component import getMultiAdapter

from DateTime import DateTime

from ftw.journal.interfaces import IJournalizer
from ftw.journal import journalMessageFactory as _


def JournalEntryHandler(event):
    """
    """
    obj = event.object
    comment = event.comment

    try:
        journalizable = IJournalizer(obj)
    except:
        return
    
    if event.action is None:
        action = _(u"label_journal_entry", default=u"Journal Entry")
    else:
        action = event.action

    if event.actor is None:
        portal_state = getMultiAdapter((obj, obj.REQUEST), name=u'plone_portal_state')
        actor = portal_state.member().getId()
    else:
        actor = event.actor
 
    if event.time is None:
        time = DateTime()
    else:
        time = event.time
    
    journalizable(action, comment, actor, time)



"""Main product initializer
"""

from zope.i18nmessageid import MessageFactory

# Define a message factory for when this product is internationalised.
# This will be imported with the special name "_" in most modules. Strings
# like _(u"message") will then be extracted by i18n tools for translation.

journalMessageFactory = MessageFactory('ftw.journal')


def initialize(context):
    """Initializer called when used as a Zope 2 product."""



from ftw.journal.interfaces import IJournalEntryEvent, IJournalizable, \
    IWorkflowHistoryJournalizable, IAnnotationsJournalizable, IJournalizer
from plone.mocktestcase import MockTestCase
from zope.interface import directlyProvides


class TestFtwJournalInterfaces(MockTestCase):

    def test_IJournalEntryEvent(self):
        """ Test event interface
        """
        self.interface_test(IJournalEntryEvent)

        # Test attributes of event
        self.assertNotEquals(IJournalEntryEvent.get('action'), None)
        self.assertNotEquals(IJournalEntryEvent.get('comment'), None)
        self.assertNotEquals(IJournalEntryEvent.get('actor'), None)
        self.assertNotEquals(IJournalEntryEvent.get('time'), None)

    def test_IJournalizable(self):
        """ Test makrer interface
        """
        self.interface_test(IJournalizable)

    def test_IWorkflowHistoryJournalizable(self):
        """ Test makrer interface
        """
        self.interface_test(IWorkflowHistoryJournalizable)

    def test_IAnnotationsJournalizable(self):
        """ Test makrer interface
        """
        self.interface_test(IAnnotationsJournalizable)

    def test_IJournalizer(self):
        """ Test makrer interface
        """
        self.interface_test(IJournalizer)

    def interface_test(self, i):
        """ Testmethod for interfaces"""
        obj = self.create_dummy()
        directlyProvides(obj, i)

        self.assertTrue(i(obj) == obj)



from ftw.journal.adapters.annotations_journal import AnnotationsJournalizable
from ftw.journal.adapters.workflow_history_journal import \
    WorkflowHistoryJournalizable
from ftw.journal.config import JOURNAL_ENTRIES_ANNOTATIONS_KEY
from mocker import ANY
from persistent.list import PersistentList
from plone.mocktestcase import MockTestCase
from Products.CMFCore.WorkflowTool import WorkflowTool
from zope.annotation.interfaces import IAnnotations
from zope.interface import Interface


class TestFtwJournalAdapters(MockTestCase):

    def test_annotations_journal(self):
        """ Test annotations journal adapter
        """
        # Context
        mock_context = self.create_dummy()

        # Annotation-Adapter
        annotation_adapter = dict()
        annotation_adapter['ftw.journal.journal_entries_annotations_key'] = []
        mock_annotation_adapter = self.mocker.proxy(
            annotation_adapter, spec=None)
        self.expect(
            mock_annotation_adapter(mock_context)).call(
                lambda a: mock_annotation_adapter).count(0, None)
        self.expect(
            mock_annotation_adapter.get(ANY, ANY)).result(PersistentList())
        self.mock_adapter(mock_annotation_adapter, IAnnotations, [Interface])


        self.replay()

        adapter = AnnotationsJournalizable(mock_context)
        adapter('save', 'saved new file', 'ratman', '15:18', )

        # Checks
        values = mock_annotation_adapter.get(
            JOURNAL_ENTRIES_ANNOTATIONS_KEY)[0].values()
        self.assertIn('save', values)
        self.assertIn('saved new file', values)
        self.assertIn('ratman', values)
        self.assertIn('15:18', values)

    def test_workflow_history_journal(self):
        """ Test workflow history journal adapter
        """
        # Context
        mock_context = self.create_dummy()

        # Workflow
        mock_wf = self.mocker.mock()
        self.expect(mock_wf.id).result('wf-published')

        # Workflowtool
        wftool = WorkflowTool()
        mock_wftool = self.mocker.proxy(wftool, spec=None)
        self.expect(mock_wftool.getWorkflowsFor(ANY)).result([mock_wf])
        self.expect(mock_wftool.getInfoFor(ANY, ANY, ANY)).result('published')

        self.mock_tool(mock_wftool, 'portal_workflow')

        self.replay()

        adapter = WorkflowHistoryJournalizable(mock_context)
        adapter('save', 'saved new file', 'ratman', '15:18', )

        # Get the saved values
        values = mock_context.workflow_history.get('wf-published')[0].values()

        # Checks
        self.assertIn('save', values)
        self.assertIn('published', values)
        self.assertIn('saved new file', values)
        self.assertIn('ratman', values)
        self.assertIn('15:18', values)



from OFS.SimpleItem import SimpleItem
from ftw.journal import journalMessageFactory as _
from ftw.journal.events.events import JournalEntryEvent
from plone.app.contentrules.browser.formhelper import AddForm, EditForm
from plone.contentrules.rule.interfaces import IRuleElementData, IExecutable
from zope import schema
from zope.component import adapts
from zope.component.interfaces import IObjectEvent
from zope.event import notify
from zope.formlib import form
from zope.interface import Interface, implements


class IJournalAction(Interface):
    """Definition of the configuration available for a mail action
    """
    action = schema.TextLine(title=_(u"Action"),
                             description=_(u"A title that will be used in the 'action' field of the journal entry."),
                             required=True)
    comment = schema.TextLine(title=_(u"Comment"),
                              description=_(u"A description that will be used in the 'comment' field in the journal entry."),
                              required=True)

class JournalAction(SimpleItem):
    """
    The implementation of the action defined before
    """
    implements(IJournalAction, IRuleElementData)

    action = u''
    comment = u''

    element = 'ftw.journal.JournalAction'

    @property
    def summary(self):
        return _(u"Journal action ${action}", mapping={'action' : self.action})

class JournalActionExecutor(object):
    """The executor for this action.
    """
    implements(IExecutable)
    adapts(Interface, IJournalAction, IObjectEvent)

    def __init__(self, context, element, event):
        self.context = context
        self.element = element
        self.event = event

    def __call__(self):
        action = self.element.action
        comment = self.element.comment
        obj = self.event.object

        notify(JournalEntryEvent(obj, comment, action))

        return True

class JournalAddForm(AddForm):
    """
    An add form for the journal action
    """
    form_fields = form.FormFields(IJournalAction)
    label = _(u"Add Journal Action")
    description = _(u"A journal action makes an entry into the journal of an object.")
    form_name = _(u"Configure element")

    def create(self, data):
        a = JournalAction()
        form.applyChanges(a, self.form_fields, data)
        return a

class JournalEditForm(EditForm):
    """
    An edit form for the journal action
    """
    form_fields = form.FormFields(IJournalAction)
    label = _(u"Edit Journal Action")
    description = _(u"A journal action makes an entry into the journal of an object.")
    form_name = _(u"Configure element")



from Acquisition import aq_inner
from ftw.journal.config import JOURNAL_ENTRIES_ANNOTATIONS_KEY
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from plone.protect import utils
from zope.annotation.interfaces import IAnnotations, IAnnotatable
from zope.interface import alsoProvides


class AnnotationsJournalizable(object):
    """Adapter to create journal entries in an annotation"""

    def __init__(self, context):
        self.context = aq_inner(context)
        alsoProvides(self.context, IAnnotatable)

    def __call__(self, action, comment, actor, time):
        context = self.context
        annotations = IAnnotations(context)
        journal_annotations = annotations.get(JOURNAL_ENTRIES_ANNOTATIONS_KEY,
                                              None)

        if not journal_annotations:
            annotations[JOURNAL_ENTRIES_ANNOTATIONS_KEY] = PersistentList()
            journal_annotations = annotations.get(
                JOURNAL_ENTRIES_ANNOTATIONS_KEY)

        history_entry = PersistentDict({
                         'action': action,
                         'comments': comment,
                         'actor': actor,
                         'time': time,
                         })

        # If we have plone.protect > 3.0, mark the journal write as safe
        if 'safeWrite' in dir(utils):
            utils.safeWrite(context)
            utils.safeWrite(journal_annotations)

        journal_annotations.append(history_entry)



from zope.component.interfaces import IObjectEvent
from zope.interface import Interface, Attribute


class IJournalEntryEvent(IObjectEvent):
    """An event that can be fired to make a journal entry
    """

    action = Attribute("")
    comment = Attribute("")
    actor = Attribute("")
    time = Attribute("")


class IJournalizable(Interface):
    """Marker interface for objects that can have journal entries set on themselves
    """


class IWorkflowHistoryJournalizable(IJournalizable):
    """Marker interface for objects that store journal entries in the workflow history
    """


class IAnnotationsJournalizable(IJournalizable):
    """Marker interface for objects that store journal entries in an annotation
    """


class IJournalizer(Interface):
    """Interface for
    """



from setuptools import setup, find_packages
import os

version = '1.3.0'
maintainer = 'Jonas Baumann'

tests_require = [
    'mocker',
    'unittest2',
    'zope.configuration',
    'plone.app.testing',
    'plone.mocktestcase',
    'plone.app.contenttypes',
    ]

setup(name='ftw.journal',
      version=version,
      maintainer=maintainer,
      description="Journaling infrastructure for plone",

      long_description=open("README.rst").read() + "\n" + \
          open(os.path.join("docs", "HISTORY.txt")).read(),

      # Get more strings from
      # http://www.python.org/pypi?%3Aaction=list_classifiers

      classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.3',
        'Framework :: Plone :: 5.1',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],

      keywords='plone workflow journal',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.journal',

      license='GPL2',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
        'setuptools',

        # Zope
        'Acquisition',
        'zope.annotation',
        'zope.component',
        'zope.event',
        'zope.formlib',
        'zope.i18nmessageid',
        'zope.interface',
        'zope.schema',
        'ZODB3',
        'Zope2',

        # Plone
        'Plone',
        'Products.CMFCore',
        'plone.app.contentrules',
        'plone.contentrules',
        ],

      tests_require=tests_require,
      extras_require = dict(tests=tests_require),

      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



from pkg_resources import get_distribution
from plone.app.testing import applyProfile
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import PLONE_FIXTURE
from zope.configuration import xmlconfig

IS_PLONE_5 = get_distribution('Plone').version >= '5'

class JournalFunctionalLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, )

    def setUpZope(self, app, configurationContext):
        from ftw import journal
        xmlconfig.file('configure.zcml', package=journal,
            context=configurationContext)
        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <include package="z3c.autoinclude" file="meta.zcml" />'
            '  <includePlugins package="plone" />'
            '  <includePluginsOverrides package="plone" />'
            '</configure>',
            context=configurationContext)

    def setUpPloneSite(self, portal):

        from plone.app.testing import setRoles, TEST_USER_ID
        setRoles(portal, TEST_USER_ID, ['Member', 'Contributor', 'Editor'])
        if IS_PLONE_5:
            applyProfile(portal, 'plone.app.contenttypes:default')

FTW_JOURNAL_FIXTURE = JournalFunctionalLayer()
FTW_JORUNAL_INTEGRATION_TESTING = IntegrationTesting(
    bases=(FTW_JOURNAL_FIXTURE, ), name="FtwJournal:Integration")



from ftw.journal.events.events import JournalEntryEvent
from ftw.journal.events.handlers import JournalEntryHandler
from ftw.journal.interfaces import IJournalizer
from plone.mocktestcase import MockTestCase
from zope.interface import Interface
from mocker import ANY
from DateTime import DateTime


class TestFtwJournalEvents(MockTestCase):

    def test_event(self):
        """ Test the events module
        """
        # Context
        mock_context = self.create_dummy()

        event = JournalEntryEvent(
            mock_context,
            'new entry',
            'save',
            'ratman',
            '15:18', )

        # Checks
        self.assertTrue(mock_context ==  event.object)
        self.assertIn('save', event.action)
        self.assertIn('new entry', event.comment)
        self.assertIn('ratman', event.actor)
        self.assertIn('15:18', event.time)

    def test_handler_not_journalizable(self):
        """ Test the handler module with no provided journalizable adapter
        """
        values = self.handler_base(journalizable=False)
        self.assertTrue(values == {})

    def test_handler_with_attrs(self):
        """ Test the handler module with all attributes"""

        values = self.handler_base(
            'new entry',
            'ratman',
            '15:18',
            'save', ).values()

        self.assertIn('save', values)
        self.assertIn('new entry', values)
        self.assertIn('ratman', values)
        self.assertIn('15:18', values)

    def test_handler_empty_attrs(self):
        """ Test the handler module with empty attrs
        """

        values = self.handler_base()

        # We just have to check the date
        time = values.get('time').Date()
        values = values.values()

        self.assertIn('label_journal_entry', values)
        self.assertIn('ratman', values)
        self.assertIn(DateTime().Date(), time)

    def handler_base(
        self,
        comment='',
        actor=None,
        time=None,
        action=None,
        journalizable=True):
        """ Basetest for the handler module
        """

        class JournalizableAdapter(object):
            """ This Class we use to store the data because we
            don't want to test the adapters and can't mock the storing
            """
            values = {}

            def __call__(self, action, comment, actor, time):
                self.values = {'action': action,
                               'comment': comment,
                               'actor': actor,
                               'time': time, }

        # Context
        mock_context = self.create_dummy()
        mock_context.REQUEST = 'portal_state'

        # Portal State
        mock_portal_state = self.mocker.mock(count=False)
        self.expect(
            mock_portal_state(ANY, mock_context.REQUEST)).result(
                mock_portal_state)
        self.expect(mock_portal_state.member().getId()).result(
            'ratman')

        self.mock_adapter(mock_portal_state, Interface, [Interface, Interface], 'plone_portal_state')

        # Adapter for IJournalizer
        mock_journalizable = self.mocker.proxy(
            JournalizableAdapter(), spec=False)
        self.expect(
            mock_journalizable(mock_context)).result(
                mock_journalizable).count(0, None)

        if journalizable:
            self.mock_adapter(mock_journalizable, IJournalizer, [Interface])

        # Event
        mock_event = self.mocker.mock(count=False)
        self.expect(mock_event.object).result(mock_context)
        self.expect(mock_event.comment).result(comment)
        self.expect(mock_event.actor).result(actor)
        self.expect(mock_event.time).result(time)
        self.expect(mock_event.action).result(action)

        self.replay()

        # Call function
        JournalEntryHandler(mock_event)

        # Return the values
        return mock_journalizable.values